using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Collections.Generic;

namespace SemiColinGames {
  class Player {
    enum Facing { Left, Right };
    enum Pose { Walking, Standing, Crouching, Stretching, SwordSwing, Jumping };

    private const int moveSpeed = 180;
    private const int jumpSpeed = -600;
    private const int gravity = 2400;

    private Texture2D texture;
    private const int spriteSize = 48;
    private const int spriteWidth = 7;

    private Point position = new Point(64, 16 * 14);
    private int jumps = 0;
    private Facing facing = Facing.Right;
    private Pose pose = Pose.Jumping;
    private double swordSwingTime = 0;
    private double jumpTime = 0;
    private float ySpeed = 0;
    private float totalModelTime = 0;

    public Player(Texture2D texture) {
      this.texture = texture;
    }

    public Point Position { get { return position; } }

    private Rectangle Bbox(Point position) {
      return new Rectangle(position.X - spriteWidth, position.Y - 7, spriteWidth * 2, 26);
    }

    public void Update(float modelTime, History<Input> input, List<Rectangle> collisionTargets) {
      totalModelTime += modelTime;
      Point oldPosition = position;
      Vector2 movement = HandleInput(modelTime, input);
      position = new Point((int) (oldPosition.X + movement.X), (int) (oldPosition.Y + movement.Y));

      Rectangle oldBbox = Bbox(oldPosition);
      Rectangle playerBbox = Bbox(position);
      bool standingOnGround = false;

      foreach (var rect in collisionTargets) {
        playerBbox = Bbox(position);

        // first we check for left-right collisions...
        if (playerBbox.Intersects(rect)) {
          if (oldBbox.Right <= rect.Left && playerBbox.Right > rect.Left) {
            position.X = rect.Left - spriteWidth;
          }
          if (oldBbox.Left >= rect.Right && playerBbox.Left < rect.Right) {
            position.X = rect.Right + spriteWidth;
          }
          playerBbox = Bbox(position);
        }
        // after fixing that, we check for hitting our head or hitting the ground.
        if (playerBbox.Intersects(rect)) {
          if (oldPosition.Y > position.Y) {
            int diff = playerBbox.Top - rect.Bottom;
            position.Y -= diff;
            // TODO: set ySpeed = 0 here so that bonking our head actually reduces hangtime?
          } else {
            standingOnGround = true;
            int diff = playerBbox.Bottom - rect.Top;
            position.Y -= diff;
          }
        } else {
          playerBbox.Height += 1;
          if (playerBbox.Intersects(rect)) {
            standingOnGround = true;
            Debug.AddRect(rect, Color.Cyan);
          } else {
            Debug.AddRect(rect, Color.Green);
          }
        }
      }
      if (standingOnGround) {
        jumps = 1;
        ySpeed = 0;
        Debug.AddRect(playerBbox, Color.Red);
      } else {
        jumps = 0;
        Debug.AddRect(playerBbox, Color.Orange);
      }

      if (movement.X > 0) {
        facing = Facing.Right;
      } else if (movement.X < 0) {
        facing = Facing.Left;
      }
      if (swordSwingTime > 0) {
        pose = Pose.SwordSwing;
      } else if (jumps == 0) {
        pose = Pose.Jumping;
      } else if (movement.X != 0) {
        pose = Pose.Walking;
      } else if (input[0].Motion.Y > 0) {
        pose = Pose.Stretching;
      } else if (input[0].Motion.Y < 0) {
        pose = Pose.Crouching;
      } else {
        pose = Pose.Standing;
      }
    }
    
    // Returns the desired (dx, dy) for the player to move this frame.
    Vector2 HandleInput(float modelTime, History<Input> input) {
      Vector2 result = new Vector2();
      result.X = (int) (input[0].Motion.X * moveSpeed * modelTime);

      if (input[0].Jump && !input[1].Jump && jumps > 0) {
        jumpTime = 0.3;
        jumps--;
        ySpeed = jumpSpeed;
      }

      if (input[0].Attack && !input[1].Attack && swordSwingTime <= 0) {
        swordSwingTime = 0.3;
      }

      result.Y = ySpeed * modelTime;
      ySpeed += gravity * modelTime;
      jumpTime -= modelTime;
      swordSwingTime -= modelTime;
      return result;
    }

    private int SpriteIndex(Pose pose) {
      int frameNum = ((int) (totalModelTime * 1000) / 125) % 4;
      if (frameNum == 3) {
        frameNum = 1;
      }
      switch (pose) {
        case Pose.Walking:
          return 6 + frameNum;
        case Pose.Stretching:
          return 18 + frameNum;
        case Pose.Jumping:
          if (jumpTime > 0.2) {
            return 15;
          } else if (jumpTime > 0.1) {
            return 16;
          } else {
            return 17;
          }
        case Pose.SwordSwing:
          if (swordSwingTime > 0.2) {
            return 30;
          } else if (swordSwingTime > 0.1) {
            return 31; 
          } else {
            return 32;
          }
        case Pose.Crouching:
          return 25;
        case Pose.Standing:
        default:
          return 7;
      }
    }

    public void Draw(SpriteBatch spriteBatch, Camera camera) {
      int index = SpriteIndex(pose);
      Rectangle textureSource = new Rectangle(index * spriteSize, 0, spriteSize, spriteSize);
      Vector2 spriteCenter = new Vector2(spriteSize / 2, spriteSize / 2);
      SpriteEffects effect = facing == Facing.Right ?
        SpriteEffects.FlipHorizontally : SpriteEffects.None;
      Vector2 drawPos = new Vector2(position.X - camera.Left, position.Y);
      spriteBatch.Draw(texture, drawPos, textureSource, Color.White, 0f, spriteCenter,
        Vector2.One, effect, 0f);
    }
  }
}
